#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/NGLInit.h"
#include "ngl/ShaderManager.h"
#include "ngl/Util.h"
#include <QImage>
#include <ctime>

/// On the mac these are extentions, it's easier to let the
/// pre-processor deal with this than using GLEW / other code
/// as on Win and Linux it works as normal glGen

#ifdef DARWIN
  #define glGenVertexArrays glGenVertexArraysAPPLE
  #define glBindVertexArray glBindVertexArrayAPPLE
#endif


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_origX=0;
  m_origY=0;
	m_fpsTimer =startTimer(0);
	  m_fps=0;
	  m_frames=0;
	  m_timer.start();
}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}

void GLWindow::loadTexture()
{
  QImage *image = new QImage();
  bool loaded=image->load("textures/crate.bmp");
  if(loaded == true)
  {
		int width=image->width();
		int height=image->height();

    unsigned char *data = new unsigned char[ width*height*3];
    unsigned int index=0;
    QRgb colour;
    for( int y=0; y<height; ++y)
    {
      for( int x=0; x<width; ++x)
      {
        colour=image->pixel(x,y);

        data[index++]=qRed(colour);
        data[index++]=qGreen(colour);
        data[index++]=qBlue(colour);
      }
    }


  GLuint textureName;
  glGenTextures(1,&textureName);
  glBindTexture(GL_TEXTURE_2D,textureName);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,width,height,0,GL_RGB,GL_UNSIGNED_BYTE,data);
	glGenerateMipmap(GL_TEXTURE_2D); //  Allocate the mipmaps

	}
}
//----------------------------------------------------------------------------------------------------------------------
/// \brief create a cube and stuff it into a VBO on the GPU
/// @param[in] _scale a scale factor for the unit vertices
void GLWindow::createCube(
                            GLfloat _scale
                         )
{

   // vertex coords array
  GLfloat vertices[] = {
                         1, 1, 1, -1, 1, 1, -1,-1, 1,  1,-1, 1,  // v0-v1-v2-v3
                         1, 1, 1,  1,-1, 1,  1,-1,-1,  1, 1,-1,  // v0-v3-v4-v5
                         1, 1, 1,  1, 1,-1, -1, 1,-1, -1, 1, 1,  // v0-v5-v6-v1
                        -1, 1, 1, -1, 1,-1, -1,-1,-1, -1,-1, 1,  // v1-v6-v7-v2
                        -1,-1,-1,  1,-1,-1,  1,-1, 1, -1,-1, 1,  // v7-v4-v3-v2
                         1,-1,-1, -1,-1,-1, -1, 1,-1,  1, 1,-1   // v4-v7-v6-v5
                        };

  GLfloat texture[] = {
                           0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
													0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
													0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
													0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
													0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
													0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3

											};



  // first we scale our vertices to _scale
  for(int i=0; i<24*3; ++i)
  {
    vertices[i]*=_scale;
  }



	// first we create a vertex array Object
  glGenVertexArrays(1, &m_vaoID);

	// now bind this to be the currently active one
  glBindVertexArray(m_vaoID);
	// now we create two VBO's one for each of the objects these are only used here
	// as they will be associated with the vertex array object
  GLuint vboID[2];
	glGenBuffers(2, &vboID[0]);
	// now we will bind an array buffer to the first one and load the data for the verts
	glBindBuffer(GL_ARRAY_BUFFER, vboID[0]);
	glBufferData(GL_ARRAY_BUFFER, 48*3*sizeof(GLfloat), vertices, GL_STATIC_DRAW);
	// now we bind the vertex attribute pointer for this object in this case the
	// vertex data
	ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  (*shader)["TextureShader"]->vertexAttribPointer("inVert",3,GL_FLOAT,0,0);
  (*shader)["TextureShader"]->enableAttribArray("inVert");

	// now we repeat for the UV data using the second VBO
  glBindBuffer(GL_ARRAY_BUFFER, vboID[1]);
	glBufferData(GL_ARRAY_BUFFER, 48*2*sizeof(GLfloat), texture, GL_STATIC_DRAW);
	// now bind
  (*shader)["TextureShader"]->vertexAttribPointer("inUV",2,GL_FLOAT,0,0);
  (*shader)["TextureShader"]->enableAttribArray("inUV");
	// finally switch back to the default so we don't overwrite
	glEnableVertexAttribArray(0);
  glBindVertexArray(0);

}




// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(2,1,2);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
   	glewInit();
   #endif
   Init->initGlew();
   m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720.0/576.0,0.5,150,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders

	shader->createShaderProgram("TextureShader");

	shader->attachShader("SimpleVertex",ngl::VERTEX);
	shader->attachShader("SimpleFragment",ngl::FRAGMENT);
	shader->loadShaderSource("SimpleVertex","shaders/Vertex.vs");
	shader->loadShaderSource("SimpleFragment","shaders/Fragment.fs");

	shader->compileShader("SimpleVertex");
	shader->compileShader("SimpleFragment");
	shader->attachShaderToProgram("TextureShader","SimpleVertex");
	shader->attachShaderToProgram("TextureShader","SimpleFragment");

	shader->bindAttribute("TextureShader",0,"inVert");
	shader->bindAttribute("TextureShader",1,"inUV");

	shader->linkProgramObject("TextureShader");
	shader->use("TextureShader");

   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("TextureShader","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
	 createCube(0.2);
	 loadTexture();
   glEnable(GL_DEPTH_TEST); // for removal of hidden surfaces
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	(*shader)["TextureShader"]->use();
  m_transformStack.loadGlobalAndCurrentMatrixToShader("TextureShader","ModelMatrix");

	shader->setShaderParamFromMatrix("TextureShader","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	// now we bind back our vertex array object and draw
  glBindVertexArray(m_vaoID);		// select first VAO

	int instances=0;
	for (float z=-14; z<15; z+=0.5)
	  {
	  for (float x=-14; x<15; x+=0.5)
	  {
	  m_transformStack.pushTransform();
	  {
	    m_transformStack.getCurrentTransform().setRotation(x*20,(x*z)*40,z*2);

	     m_transformStack.getCurrentTransform().setPosition(ngl::Vector(x,0.49,z));
	     shader->setShaderParamFromMatrix("TextureShader","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
			 ++instances;
			 glDrawArrays(GL_QUADS, 0, 24);	// draw first object
	   } // and before a pop
	   m_transformStack.popTransform();

	 }
	}
	// go back to default just incase
	glEnableVertexAttribArray(0);
  glBindVertexArray(0);
	// calculate and draw FPS
	++m_frames;
	glUseProgramObjectARB(0);
	glColor3f(1,1,1);
	QFont font;
	font.setBold(true);
	font.setPointSize(20);
	QString text=QString("Texture and Vertex Array Object %1 instances Demo %2 fps").arg(instances).arg(m_fps);
	//renderText(10,20,text,font);
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	if(_event->timerId() == m_fpsTimer)
	  {
	    if( m_timer.elapsed() > 1000.0)
	    {
	      m_fps=m_frames;
	      m_frames=0;
	      m_timer.restart();
	    }
	   }
	    // re-draw GL
	updateGL();
}

